#!/usr/bin/env python
# PYTHON_ARGCOMPLETE_OK
from . import main, calib, audio, async
from .config import bitrates
from . import version

import os
import sys
import zlib
import logging
import argparse

# Python 3 has `buffer` attribute for byte-based I/O
_stdin = getattr(sys.stdin, 'buffer', sys.stdin)
_stdout = getattr(sys.stdout, 'buffer', sys.stdout)


try:
    import argcomplete
except ImportError:
    argcomplete = None

log = logging.getLogger('__name__')

bitrate = os.environ.get('BITRATE', 1)
config = bitrates.get(int(bitrate))


class Compressor(object):
    def __init__(self, stream):
        self.obj = zlib.compressobj()
        log.info('Using zlib compressor')
        self.stream = stream

    def read(self, size):
        while True:
            data = self.stream.read(size)
            if data:
                result = self.obj.compress(data)
                if not result:  # compression is too good :)
                    continue  # try again (since falsy data = EOF)
            elif self.obj:
                result = self.obj.flush()
                self.obj = None
            else:
                result = ''  # EOF marker
            return result


class Decompressor(object):
    def __init__(self, stream):
        self.obj = zlib.decompressobj()
        log.info('Using zlib decompressor')
        self.stream = stream

    def write(self, data):
        self.stream.write(self.obj.decompress(bytes(data)))

    def flush(self):
        self.stream.write(self.obj.flush())


def FileType(mode, interface_factory=None):
    def opener(fname):
        audio_interface = interface_factory() if interface_factory else None

        assert 'r' in mode or 'w' in mode
        if audio_interface is None and fname is None:
            fname = '-'

        if fname is None:
            assert audio_interface is not None
            if 'r' in mode:
                s = audio_interface.recorder()
                return async.AsyncReader(stream=s, bufsize=s.bufsize)
            if 'w' in mode:
                return audio_interface.player()

        if fname == '-':
            if 'r' in mode:
                return _stdin
            if 'w' in mode:
                return _stdout

        return open(fname, mode)

    return opener


def get_volume_cmd(args):
    volume_controllers = [
        dict(test='pactl --version',
             send='pactl set-sink-volume @DEFAULT_SINK@',
             recv='pactl set-source-volume @DEFAULT_SOURCE@')
    ]
    if args.calibrate == 'auto':
        for c in volume_controllers:
            if os.system(c['test']) == 0:
                return c[args.command]


def wrap(cls, stream, enable):
    return cls(stream) if enable else stream


def create_parser(description, interface_factory):
    p = argparse.ArgumentParser(description=description)
    subparsers = p.add_subparsers()

    # Modulator
    sender = subparsers.add_parser(
        'send', help='modulate binary data into audio signal.')
    sender.add_argument(
        '-i', '--input', help='input file (use "-" for stdin).')
    sender.add_argument(
        '-o', '--output', help='output file (use "-" for stdout).'
        ' if not specified, `aplay` tool will be used.')
    sender.add_argument(
        '-g', '--gain', type=float, default=1.0,
        help='Modulator gain (defaults to 1)')
    sender.set_defaults(
        main=lambda config, args: main.send(
            config, src=wrap(Compressor, args.src, args.zlib), dst=args.dst,
            gain=args.gain
        ),
        calib=lambda config, args: calib.send(
            config=config, dst=args.dst,
            volume_cmd=get_volume_cmd(args)
        ),
        input_type=FileType('rb'),
        output_type=FileType('wb', interface_factory),
        command='send'
    )

    # Demodulator
    receiver = subparsers.add_parser(
        'recv', help='demodulate audio signal into binary data.')
    receiver.add_argument(
        '-i', '--input', help='input file (use "-" for stdin).'
        ' if not specified, `arecord` tool will be used.')
    receiver.add_argument(
        '-o', '--output', help='output file (use "-" for stdout).')
    receiver.add_argument(
        '-d', '--dump', type=FileType('wb'),
        help='Filename to save recorded audio')
    receiver.add_argument(
        '--plot', action='store_true', default=False,
        help='plot results using pylab module')
    receiver.set_defaults(
        main=lambda config, args: main.recv(
            config, src=args.src, dst=wrap(Decompressor, args.dst, args.zlib),
            pylab=args.pylab, dump_audio=args.dump
        ),
        calib=lambda config, args: calib.recv(
            config=config, src=args.src, verbose=args.verbose,
            volume_cmd=get_volume_cmd(args)
        ),
        input_type=FileType('rb', interface_factory),
        output_type=FileType('wb'),
        command='recv'
    )

    calibration_help = ('Run calibration '
                        '(specify "auto" for automatic gain control)')

    for sub in subparsers.choices.values():
        sub.add_argument('-c', '--calibrate', nargs='?', default=False,
                         metavar='SYSTEM', help=calibration_help)
        sub.add_argument('-l', '--audio-library', default='libportaudio.so',
                         help='File name of PortAudio shared library.')
        sub.add_argument('-z', '--zlib', default=False, action='store_true',
                         help='Use zlib to compress/decompress data.')
        g = sub.add_mutually_exclusive_group()
        g.add_argument('-v', '--verbose', default=0, action='count')
        g.add_argument('-q', '--quiet', default=False, action='store_true')

    if argcomplete:
        argcomplete.autocomplete(p)

    return p


class _Dummy(object):
    def __enter__(self):
        return self

    def __exit__(self, *args):
        pass


def _main():
    fmt = ('Audio OFDM MODEM v{0:s}: '
           '{1:.1f} kb/s ({2:d}-QAM x {3:d} carriers) '
           'Fs={4:.1f} kHz')
    description = fmt.format(version.__doc__,
                             config.modem_bps / 1e3, len(config.symbols),
                             config.Nfreq, config.Fs / 1e3)
    interface = None

    def interface_factory():
        return interface

    p = create_parser(description, interface_factory)

    args = p.parse_args()
    if args.verbose == 0:
        level, fmt = 'INFO', '%(message)s'
    elif args.verbose == 1:
        level, fmt = 'DEBUG', '%(message)s'
    elif args.verbose >= 2:
        level, fmt = ('DEBUG', '%(asctime)s %(levelname)-10s '
                               '%(message)-100s '
                               '%(filename)s:%(lineno)d')
    if args.quiet:
        level, fmt = 'WARNING', '%(message)s'
    logging.basicConfig(level=level, format=fmt)

    # Parsing and execution
    log.debug(description)

    args.pylab = None
    if getattr(args, 'plot', False):
        import pylab  # pylint: disable=import-error
        args.pylab = pylab

    if args.audio_library == 'ALSA':
        from . import alsa
        interface = alsa.Interface(config)
    elif args.audio_library == '-':
        interface = _Dummy()
    else:
        interface = audio.Interface(config)
        interface.load(args.audio_library)

    with interface:
        args.src = args.input_type(args.input)
        args.dst = args.output_type(args.output)
        try:
            if args.calibrate is False:
                args.main(config=config, args=args)
            else:
                args.calib(config=config, args=args)
        finally:
            args.src.close()
            args.dst.close()
            log.debug('Finished I/O')


if __name__ == '__main__':
    _main()



from . import common
from . import dsp
from . import sampling
from . import stream

import numpy as np
import itertools
import logging
import subprocess

log = logging.getLogger(__name__)


def volume_controller(cmd):
    def controller(level):
        assert 0 < level <= 1
        percent = 100 * level
        args = '{0} {1:.0f}%'.format(cmd, percent)
        log.debug('Setting volume to %7.3f%% -> "%s"', percent, args)
        subprocess.check_call(args=args, shell=True)
    return controller if cmd else (lambda level: None)


def send(config, dst, volume_cmd=None, gain=1.0, limit=None):
    volume_ctl = volume_controller(volume_cmd)
    volume_ctl(1.0)  # full scale output volume

    calibration_symbols = int(1.0 * config.Fs)
    t = np.arange(0, calibration_symbols) * config.Ts
    signals = [gain * np.sin(2 * np.pi * f * t) for f in config.frequencies]
    signals = [common.dumps(s) for s in signals]

    for signal in itertools.islice(itertools.cycle(signals), limit):
        dst.write(signal)


def frame_iter(config, src, frame_length):
    frame_size = frame_length * config.Nsym * config.sample_size
    omegas = 2 * np.pi * np.array(config.frequencies) / config.Fs

    while True:
        data = src.read(frame_size)
        if len(data) < frame_size:
            return
        data = common.loads(data)
        frame = data - np.mean(data)

        sampler = sampling.Sampler(frame)
        symbols = dsp.Demux(sampler, omegas, config.Nsym)

        symbols = np.array(list(symbols))
        coeffs = np.mean(np.abs(symbols) ** 2, axis=0) ** 0.5

        peak = np.max(np.abs(frame))
        total = np.sqrt(np.dot(frame, frame) / (0.5 * len(frame)))
        yield coeffs, peak, total


def detector(config, src, frame_length=200):

    errors = ['weak', 'strong', 'noisy']
    for coeffs, peak, total in frame_iter(config, src, frame_length):
        max_index = np.argmax(coeffs)
        freq = config.frequencies[max_index]
        rms = abs(coeffs[max_index])
        coherency = rms / total
        flags = [total > 0.1, peak < 1.0, coherency > 0.99]

        success = all(flags)
        if success:
            msg = 'good signal'
        else:
            msg = 'too {0} signal'.format(errors[flags.index(False)])

        yield common.AttributeHolder(dict(
            freq=freq, rms=rms, peak=peak, coherency=coherency,
            total=total, success=success, msg=msg
        ))


def volume_calibration(result_iterator, volume_ctl):
    min_level = 0.01
    max_level = 1.0
    level = 0.5
    step = 0.25

    target_level = 0.4  # not too strong, not too weak
    iters_per_update = 10  # update every 2 seconds

    for index, result in enumerate(itertools.chain([None], result_iterator)):
        if index % iters_per_update == 0:
            if index > 0:  # skip dummy (first result)
                sign = 1 if (result.total < target_level) else -1
                level = level + step * sign
                level = min(max(level, min_level), max_level)
                step = step * 0.5

            volume_ctl(level)  # should run "before" first actual iteration

        if index > 0:  # skip dummy (first result)
            yield result


def iter_window(iterable, size):
    block = []
    while True:
        item = next(iterable)
        block.append(item)
        block = block[-size:]
        if len(block) == size:
            yield block


def recv(config, src, verbose=False, volume_cmd=None, dump_audio=None):
    fmt = '{0.freq:6.0f} Hz: {0.msg:20s}'
    if verbose:
        fields = ['total', 'rms', 'coherency', 'peak']
        fmt += ', '.join('{0}={{0.{0}:.4f}}'.format(f) for f in fields)

    volume_ctl = volume_controller(volume_cmd)

    if dump_audio:
        src = stream.Dumper(src, dump_audio)
    result_iterator = detector(config=config, src=src)
    result_iterator = volume_calibration(result_iterator, volume_ctl)
    for _prev, curr, _next in iter_window(result_iterator, size=3):
        # don't log errors during frequency changes
        if _prev.success and _next.success and _prev.freq != _next.freq:
            curr.msg = curr.msg if curr.success else 'frequency change'
        log.info(fmt.format(curr))



import threading
import six  # since `Queue` module was renamed to `queue` (in Python 3)
import logging

log = logging.getLogger()


class AsyncReader(object):
    def __init__(self, stream, bufsize):
        self.stream = stream
        self.queue = six.moves.queue.Queue()
        self.stop = threading.Event()
        args = (stream, bufsize, self.queue, self.stop)
        self.thread = threading.Thread(target=AsyncReader._thread,
                                       args=args, name='AsyncReader')
        self.thread.start()
        self.buf = b''

    @staticmethod
    def _thread(src, bufsize, queue, stop):
        total = 0
        try:
            log.debug('AsyncReader thread started')
            while not stop.isSet():
                buf = src.read(bufsize)
                queue.put(buf)
                total += len(buf)
            log.debug('AsyncReader thread stopped (read %d bytes)', total)
        except BaseException:
            log.exception('AsyncReader thread failed')
            queue.put(None)

    def read(self, size):
        while len(self.buf) < size:
            buf = self.queue.get()
            if buf is None:
                raise IOError('cannot read from stream')
            self.buf += buf

        result = self.buf[:size]
        self.buf = self.buf[size:]
        return result

    def close(self):
        if self.stream is not None:
            self.stop.set()
            self.thread.join()
            self.stream.close()
            self.stream = None



#!/usr/bin/env python
from setuptools import setup
from setuptools.command.test import test as TestCommand

import os
import ast

def parse_vesrion():
    cwd = os.path.dirname(__name__)
    version_file = os.path.join(cwd, 'amodem', 'version.py')

    tree = ast.parse(open(version_file).read())
    expr, = tree.body
    return expr.value.s

class PyTest(TestCommand):

    def finalize_options(self):
        self.test_args = []
        self.test_suite = True

    def run_tests(self):
        import sys
        import pytest
        sys.exit(pytest.main(['tests']))

setup(
    name='amodem',
    version=parse_vesrion(),
    description='Audio Modem Communication Library',
    author='Roman Zeyde',
    author_email='roman.zeyde@gmail.com',
    license='MIT',
    url='http://github.com/romanz/amodem',
    packages=['amodem'],
    tests_require=['pytest'],
    cmdclass={'test': PyTest},
    install_requires=['numpy', 'six'],
    platforms=['POSIX'],
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Intended Audience :: Information Technology',
        'License :: OSI Approved :: MIT License',
        'Operating System :: POSIX',
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.2',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: System :: Networking',
        'Topic :: Communications',
    ],
    entry_points={'console_scripts': ['amodem = amodem.__main__:_main']},
)



import subprocess
import logging

log = logging.getLogger(__name__)


class Interface(object):

    RECORDER = 'arecord'
    PLAYER = 'aplay'

    def __init__(self, config):
        self.config = config
        rate = int(config.Fs)
        bits_per_sample = config.bits_per_sample
        assert bits_per_sample == 16

        args = '-f S{0:d}_LE -c 1 -r {1:d} -T 100 -q -'
        args = args.format(bits_per_sample, rate).split()

        self.record_cmd = [self.RECORDER] + args
        self.play_cmd = [self.PLAYER] + args
        self.processes = []

    def __enter__(self):
        return self

    def __exit__(self, *args):
        for p in self.processes:
            try:
                p.wait()
            except OSError:
                log.warning('%s failed', p)

    def launch(self, **kwargs):
        log.debug('Launching subprocess: %s', kwargs)
        p = subprocess.Popen(**kwargs)
        self.processes.append(p)
        return p

    def recorder(self):
        return Recorder(self)

    def player(self):
        return Player(self)


class Recorder(object):
    def __init__(self, lib):
        self.p = lib.launch(args=lib.record_cmd, stdout=subprocess.PIPE)
        self.read = self.p.stdout.read
        self.bufsize = 4096

    def close(self):
        self.p.kill()


class Player(object):
    def __init__(self, lib):
        self.p = lib.launch(args=lib.play_cmd, stdin=subprocess.PIPE)
        self.write = self.p.stdin.write

    def close(self):
        self.p.stdin.close()
        self.p.wait()
